#!/bin/bash

############
# SETTINGS #

# if you have compiled the "wav2pbm" C program
# you may want to use this - it is MUCH MUCH MUCH faster than python wav2png
#WAV_TO_IMG_EXE="./wav2pbm_c/wav2pbm"
#STRIP_IMG_EXT=".pgm"

WAV_TO_IMG_EXE="python wav2png.py"
STRIP_IMG_EXT=".png"

# path where image strips will be stored
STRIPS_PATH="./strips"

# Contact Sheet images horizontally
STRIPS_PER_PAGE_X="30"
# Contact Sheet images vertically
STRIPS_PER_PAGE_Y="1"
# Horizontal Spacing
PAGE_SPACING_X="0"
# Vertical Spacing
PAGE_SPACING_Y="0"
# Contact Sheet Directory
PAGEPATH="contact"
# temporary converted unsigned 8bit audio
TEMP_AUDIO="tmp-pcm_u8.wav"

############

# ----------------------------------------------
echo
echo "Converting audio to image strips"

# what input sound file do we want to convert?
read -e -p "source audio filename: " -i "./dm.m4a" SOURCE_AUDIO

# how many dpi will we be printing/scanning?
#read -e -p "dots per inch: " -i "600" STRIP_IMG_DPI
STRIP_IMG_DPI="300"

# image wildcard 
STRIP_IMG_WILDCARD="*${STRIP_IMG_EXT}"
# per-strip image width
STRIP_WIDTH=$STRIP_IMG_DPI
# per-strip image height
let STRIP_HEIGHT=$(( $STRIP_IMG_DPI * 11 ))
# Contact sheet each-item size
STRIPSIZE="${STRIP_WIDTH}x${STRIP_HEIGHT}"

echo
echo "pre-processing audio"

# -S, show progress
# highpass 170, remove low rumble
# compand, heavy compression
# gain -h bass treble treble, compensation for paper/printer frequency response
# rate strip_height, simplify image conversion - 1s of audio = strip_height pixels
# gain -n, normalise
avconv -i "$SOURCE_AUDIO" \
-ar 22050 \
-ac 1 \
-acodec pcm_u8 \
-f u8 - | \
sox -S \
-t u8 \
-r 22050 \
- "$TEMP_AUDIO" \
highpass 170 \
compand 0.1,1 -60,-20,-6,0 \
gain -h treble 9 4000 treble 18 8000 treble 24 11000 \
compand 0.1,1 -60,-20,-6,0 \
rate $STRIP_HEIGHT \
gain -n

# make a directory for image strips
mkdir -p "$STRIPS_PATH"

# apply the conversion (python or c-based)
$WAV_TO_IMG_EXE \
-i "${STRIPS_PATH}/_" \
-w "$TEMP_AUDIO" \
-d $STRIP_WIDTH 

# remove the temporary pcm_u8 audio file
#rm "$TEMP_AUDIO"

# ----------------------------------------------
echo
echo "Step 2 - combining image strips to contact sheets for printing"
#    NOTE - time runs top to bottom, then left to right
#    at the top left of page 1 is time=0, bottom left is time=1s, etc

# Modified digital camera thumbnail/contact sheet tool.
# Original by John Hobbs
# http://www.velvetcache.org/2009/03/30/imagemagick-thumbnails-and-contact-sheets
# http://github.com/jmhobbs/helper-scripts
#
# This version modified by zenpho 2014
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
 
# CHANGELOG
# 2014-06-01 - zenpho "opticalAudio"
# 2014-03-04 - Modified to run from a "main project" directory
# 2010-06-14 - Fixed contact sheet problem, thanks to Glenn Turnbull. (John Hobbs)
# 2009-03-30 - Created script. (John Hobbs)

# change to (image strips) working directory
cd "$STRIPS_PATH"
CONTACTCOUNT=$(($STRIPS_PER_PAGE_X * $STRIPS_PER_PAGE_Y))
PIX=$(ls -l ${STRIP_IMG_WILDCARD} | wc -l)

echo
echo "Creating Contact Sheets from $PIX Images"
 
mkdir -p $PAGEPATH
CTR=0
PAGES=$(($PIX / $CONTACTCOUNT))
if [ $(($PIX % $CONTACTCOUNT)) -ne 0 ]; then
    PAGES=$(($PAGES + 1))
fi
 
PAGE=1
LIST=""
for i in $STRIP_IMG_WILDCARD; do
    if [ $(($CTR % $CONTACTCOUNT)) -eq 0 ] && [ $CTR -ne 0 ]; then
	printf -v ZEROPADPAGE "%04d" $PAGE
        echo "page $ZEROPADPAGE of $PAGES"
        convert -monitor -colorspace Gray $LIST +append ${PAGEPATH}/${ZEROPADPAGE}.png
        LIST=""
        PAGE=$(($PAGE + 1))
    fi
    LIST="$LIST $i"
    CTR=$(($CTR + 1))
done
 
if [ "" != "$LIST" ]; then
    printf -v ZEROPADPAGE "%04d" $PAGE
    echo "page $ZEROPADPAGE of $PAGES"
    convert -monitor -colorspace Gray $LIST +append ${PAGEPATH}/${ZEROPADPAGE}.png
fi

echo 
echo Moving $PAGEPATH to main directory
rm -r ../$PAGEPATH
mv $PAGEPATH ../

echo "Done creating contact sheets"
cd ..
  
