/*
Phill Phelps (zenpho@zenpho.co.uk) 2014 June 04

Create a wav sound file (pcm_u8) from a PBM image file (P5)
in the style of "variable area" optical audio for film sound

*/

#include <stdio.h>
#include <sndfile.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <dirent.h>

// globally show/hide debug messages
unsigned char gDebug = 0;

// configuration parameters for reading/decoding image data
typedef struct
{
  int imgWidth;
  int imgHeight;
  int imgMaxIntensity;

} t_config;

// ///////////////////////////////////////////////////////////////////

void disp_sfinfo(SF_INFO *sfinfo);
void exitDisplayingUsage(char *command);
void readPgmBinaryImgData(t_config *config, char* filename, 
                          unsigned char **imgData, int onlyProcessHeader);

void decodeImgDataToSndData(t_config *config, unsigned char *imgData,
                            unsigned char *sndData);

int filterDirEntries(const struct dirent *dirEntry);

// ///////////////////////////////////////////////////////////////////

int main(int argc, char* argv[])
{
  char *command = argv[0];
  char *imgDirName = NULL;    // input (sound)filename
  char *outfilename = NULL;   // output (image)name prefix
  char imgFilePath[256];

  t_config config;

  SNDFILE *outfile = NULL;
  SF_INFO sfinfo;

  int numImages;
  int imgHeightSum;

  // so as to avoid holding the entire wav data in memory
  // these buffers are overwritten for each image processed
  // imgData is overwritten as each new image is loaded
  // sndData is overwritten during decoding
  //   and flushed to disk afterwards
  unsigned char *sndData = NULL; 
  unsigned char *imgData = NULL;

  struct dirent **fileEntities;
  int numFileEntities;
  int fileEntIdx;

  // /////////////////////////////////////////////////////////////////

  // set default config parameters
  imgDirName = "stripsPGM/";
  outfilename = "out.wav";

  // parse command line options
  int c;
  int foundOptM, foundOptW;
  foundOptM = foundOptW = 0;
  while ((c = getopt(argc, argv, "m:w:b")) != -1)
  {
	  switch (c) 
    {
	  case 'm':
		  imgDirName = optarg;
      foundOptM = 1;
      break;
	  case 'w':
		  outfilename = optarg;
      foundOptW = 1;
      break;
    case 'b':
      gDebug = 1;
      break;
    default: // abort if something goes wrong
      exitDisplayingUsage(command);
    }
  }

  // missing required commandline options?
  if ( !(foundOptM && foundOptW) ) exitDisplayingUsage(command);

  // ///////////////////////////////////////////////////////////////////////
  
  /*
   before we open a wav file for writing - 
   we must know the sample rate and number of samples

   we scan the resolution of all images and count how many images we find
   -- each image represents a second of audio so we can
      count the total number of pixels and divide by the number of images
  */
  printf("looking in %s directory for images...\n", imgDirName);
  printf("counting images, calculating average imgHeight, thus samplerate\n");

  // abort if we fail to open the given image directory
  numFileEntities = scandir(imgDirName, &fileEntities, filterDirEntries, alphasort);
  if (numFileEntities < 0) {
    printf("error opening image directory\n");
    exit(1);
  }
  
  // iterate over the images in the directory
  // count the number of images - and the average image height
  numImages = 0;
  imgHeightSum = 0;
  fileEntIdx = numFileEntities;
  while( fileEntIdx-- )
  {
    // generate file path
    sprintf(imgFilePath, "%s/%s", imgDirName, fileEntities[fileEntIdx]->d_name);

    readPgmBinaryImgData(&config, imgFilePath, NULL, 1); // no imgData ptr, just read header
    imgHeightSum += config.imgHeight;
    numImages++;
  }

  // now we can set the sound file header parameters
  memset(&sfinfo, 0, sizeof(sfinfo));
  sfinfo.frames = imgHeightSum * numImages;
  sfinfo.samplerate = imgHeightSum / numImages;
  sfinfo.channels = 1;
  sfinfo.format = (SF_FORMAT_WAV | SF_FORMAT_PCM_U8);
  sfinfo.sections = 1;
  sfinfo.seekable = 1;

  // abort if sound parameters are invalid
  if ( !sf_format_check(&sfinfo) ) {
    disp_sfinfo(&sfinfo);
    printf("error - libsndfile says above params are invalid");
    exit(1);
  }

  // ///////////////////////////////////////////////////////////////////////

  // open sound file for writing
  outfile = sf_open(outfilename, SFM_WRITE, &sfinfo);
  if (gDebug) printf("opened %s for writing\n", outfilename);
  
  // iterate over the images in the directory
  // count the number of images - and the average image height
  numImages = 0;
  imgHeightSum = 0;
  fileEntIdx = 0;
  while( fileEntIdx < numFileEntities )
  {
      // generate file path
      sprintf(imgFilePath, "%s/%s", imgDirName, fileEntities[fileEntIdx]->d_name);     
  
      // attempt to read the pgm file
      // we will dynamically (re)alloc the imgData array
      // we call exit() if there is a problem
      // remember imgData is a pointer to the imgData pointer in main()
      readPgmBinaryImgData(&config, imgFilePath, &imgData, 0);

      // (re)allocate memory for the sound buffer
	    sndData = realloc(sndData, config.imgHeight);
      if (gDebug) printf("allocated sound buffer\n");

      // fill the sndData buffer
      decodeImgDataToSndData(&config, imgData, sndData);

      // write raw pcm_u8 data
      sf_write_raw(outfile, sndData, config.imgHeight);

      printf("processed %s ok\n", imgFilePath);
      free(fileEntities[fileEntIdx]);

      fileEntIdx++;
  }
  free(fileEntities);

  // close the sound file  
  sf_close(outfile);
  printf("finished processing %s", outfilename);

	// free heap memory and exit
	if(imgData != NULL) free(imgData);
	if(sndData != NULL) free(sndData);
  if (gDebug) printf("freed allocated memory\n");
  exit(0);
}


/*
  for debugging - display the libsndfile SFINFO struct 
*/
void disp_sfinfo(SF_INFO *sfinfo)
{
   printf("frames:\t\t%ld\n",      (long)sfinfo->frames);
   printf("samplerate:\t%d\n",     sfinfo->samplerate);
   printf("channels:\t%d\n",       sfinfo->channels);
   printf("format number:\t%d\n",  sfinfo->format);
   printf("sections:\t%d\n",       sfinfo->sections);
   printf("seekable?:\t%d\n",      sfinfo->seekable);
   printf("\n");
}


/*
  remind the user what command line arguments we support
*/
void exitDisplayingUsage(char *command)
{
   // display usage and exit with returncode 1
   printf("usage: %s -m <image directory> -w <wav filename>\n", command);
   exit(1);
}

/*
  determine PGM (P5) properties (e.g. imgWidth, imgHeight, imgMaxIntensity)
  optionally abort after populating the config struct with these

  optionally continue and fill the imgData array with pixel values

  remember imgData is itself a pointer to the imgData pointer in main()
*/
void 
readPgmBinaryImgData(
  t_config *config, 
  char* filename, 
  unsigned char **imgData,
  int onlyProcessHeader // should we abort after filling imgHeight,width,etc 
  )
{
  FILE *infile = NULL;

  // open pgm file for writing
  // abort if the output file cannot be opened for writing
  infile = fopen(filename, "rb");
  if (!infile)
  {
    printf("==Error, cannot open image file %s for reading==\n", filename);
    exit(1);
  }
  if (gDebug) printf("image file %s open for reading ok\n", filename);

  // read header - abort if not "P5" format
  char formatChar;
  int formatNum;
  fscanf(infile, "%c%d\n", &formatChar, &formatNum);
  if( formatChar != 'P' && formatChar != 5 ){
     printf("error - %s is not in required P5 format\n", filename);
     exit(1);
  }

  // eat up any comment lines which come between the P5\n and the dimensions
  int c = fgetc(infile);
  ungetc(c, infile);  
  if( c == '#' ) {
    while( fgetc(infile) != '\n' );
  }

  // read the image dimensions
  fscanf(infile, "%d %d\n", &config->imgWidth, &config->imgHeight);

  // read the maximum intensity per-pixel
  fscanf(infile, "%d", &config->imgMaxIntensity);
  
  if (gDebug) printf("image is %d by %d, range 0 to %d\n",
      config->imgWidth, 
      config->imgHeight,
      config->imgMaxIntensity
  );      

  // if we've been told to abort after reading header
  if (onlyProcessHeader){
    fclose(infile);
    return;
  }

  // dynamically (re)allocate imgData array
  // remember imgData is a pointer to the imgData pointer in main()
  *imgData = realloc(*imgData, config->imgWidth * config->imgHeight);
  if (gDebug) printf("allocated image buffer\n");

  // read raster
  // remember imgData is a pointer to the imgData pointer in main()
  int pixelsToProcess = config->imgWidth * config->imgHeight;
  fread(*imgData, 1, pixelsToProcess, infile);

  fclose(infile);
  if (gDebug) printf("finished reading pgm image ok\n");
}

/*
  fill sndData array based on dark/light pixels in imgData array
*/
void decodeImgDataToSndData(
  t_config *config,  
  unsigned char *imgData,
  unsigned char *sndData)
{
  int xSum, xAvg;
  int linesToProcess = config->imgHeight;
  int pixelsThisLineToProcess;
  while(linesToProcess--)
  {
    // calculate the avg pixel value for each x-axis row
    xSum = 0;
    pixelsThisLineToProcess = config->imgWidth;
    while(pixelsThisLineToProcess--)
    {
      xSum += *imgData;
      imgData++;
    }
    xAvg = xSum / config->imgWidth;
    
    // clip
    if (xAvg < 0) xAvg = 0;
    if (xAvg > 255) xAvg = 255;

    // TODO - high pass filter here
    
    // store sound sample
    *sndData = xAvg;
    sndData++;
  }
}

/*
  this is used by scandir() to filter 'interesting' directory entries
  
  we return 1 if we are interested in this entry
  otherwise we return 0 (i.e. we exclude this entry)

  the 'rules' to determine if a file are wanted/unwanted
  are explained as inline comments in the function body
*/
int filterDirEntries(const struct dirent *dirEntry)
{
    // if there is a dot at the beginning of the name
    if (dirEntry->d_name[0] == '.') return 0;

    // if the file is NOT a 'regular file'
    if (dirEntry->d_type != DT_REG) return 0;

    return 1;
}

