/*
Phill Phelps (zenpho@zenpho.co.uk) 2014 June 04

Create a serise of PBM images (P5) from a single wav sound file (pcm_u8)
in the style of "variable area" optical audio for film sound

*/

#include <stdio.h>
#include <sndfile.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>

// globally show/hide debug messages
unsigned char gDebug = 0;

// configuration parameters for plotting/writing image data
typedef struct
{
	int imgWidth;
	int imgHeight;

	int sndSamplerate; // from SF_INFO
  sf_count_t sndSamplesTotal; // from SF_INFO
  sf_count_t sndSamplesProcessed; // from sf_read_???()

  float gain;
} t_config;

// ///////////////////////////////////////////////////////////////////

void disp_sfinfo(SF_INFO *sfinfo);
void exitDisplayingUsage(char *command);
void calcAmpEnv(t_config *config, unsigned char *sndData, int *ampEnvelope);
void displayAmpEnv(t_config *config, int *ampEnvelope);
void plotImageData(t_config *config, int *ampEnvelope, unsigned char *imgData);
void writeImageData(t_config *config, unsigned char *imgData, char* filename);

// ///////////////////////////////////////////////////////////////////

int main(int argc, char* argv[])
{
  char *command = argv[0];
  char *infilename = NULL;    // input (sound)filename
  char *outfileprefix = NULL; // output (image)name prefix

  SNDFILE *infile = NULL;

  SF_INFO sfinfo;
  t_config config;

  int *ampEnvelope = NULL;
  unsigned char *sndData = NULL;
  unsigned char *imgData = NULL;

  // set default config parameters
	memset(&config, 0, sizeof(config));
  infilename = "input.wav";
  outfileprefix = "out";
  config.imgWidth = 200;

  // parse command line options
  int c;
  int foundOptW, foundOptI, foundOptD;
  foundOptW = foundOptI = foundOptD = 0;
  while ((c = getopt(argc, argv, "w:i:d:g:b")) != -1)
  {
		switch (c) 
    {
		case 'w':
			infilename = optarg;
      foundOptW = 1;
      break;
		case 'i':
			outfileprefix = optarg;
      foundOptI = 1;
      break;
		case 'd':
			config.imgWidth = atoi(optarg);
      foundOptD = 1;
      break;
    case 'b':
      gDebug = 1;
      break;
    default: // abort if something goes wrong
      exitDisplayingUsage(command);
    }
  }
  
  // missing required options?
  if ( !(foundOptW && foundOptI && foundOptD) ) exitDisplayingUsage(command);

  // abort if we fail to read the input file
  memset(&sfinfo, 0, sizeof(sfinfo));
  if ((infile = sf_open(infilename, SFM_READ, &sfinfo)) == NULL)
  {
    printf("==%s Error reading/parsing audio==\n", infilename);
    sf_close(infile);
    exit(1);
  }

  // abort if the file format isn't supported
  if (sfinfo.channels != 1)
  {
      printf("==%s Error, audio file is not mono==\n", infilename);
      sf_close(infile);
      exit(1);
  }

  // display the sound info
  disp_sfinfo(&sfinfo);

	// set the remaining configuration parameters
	config.imgHeight = sfinfo.samplerate;
	config.sndSamplerate = sfinfo.samplerate;
  config.sndSamplesTotal = sfinfo.frames ; // remember we only process mono
  config.sndSamplesProcessed = 0;

  // prepare the input and envelope buffers
	sndData = malloc(config.sndSamplerate);
  if (gDebug) printf("allocated pcm buffer\n");

	ampEnvelope = malloc(sizeof(int) * config.imgHeight);
  if (gDebug) printf("allocated envelope buffer\n");

  imgData = malloc(config.imgWidth * config.imgHeight);
  if (gDebug) printf("allocated image buffer\n");

  // process the entire sound file
  int imgCount = 0;
  char outfilename[80];
  do
  {
    // read one second of sound input data
    config.sndSamplesProcessed += 
        sf_read_raw(infile, sndData, config.sndSamplerate);
    if (gDebug) printf("filled sound buffer\n");

	  // compute one second of amplitude envelope
    calcAmpEnv(&config, sndData, ampEnvelope);

    // generate image data and write to disk as png
    plotImageData(&config, ampEnvelope, imgData);
    
    // format image data as pgm binary (P5) and write to disk
    snprintf(outfilename, 80, "%s%010d.pgm", outfileprefix, imgCount);
    writeImageData(&config, imgData, outfilename);

    if (gDebug) printf("read %d of %d sound samples\n", 
        (int)config.sndSamplesProcessed,
        (int)config.sndSamplesTotal);
    imgCount++;
  }while(config.sndSamplesProcessed < config.sndSamplesTotal);

	// free heap memory and exit
  if(imgData != NULL) free(imgData);
	if(ampEnvelope != NULL) free(ampEnvelope);
	if(sndData != NULL) free(sndData);
  if (gDebug) printf("freed allocated memory\n");
  exit(0);
}


/*
  for debugging - display the libsndfile SFINFO struct 
*/
void disp_sfinfo(SF_INFO *sfinfo)
{
   printf("frames:\t\t%ld\n",      (long)sfinfo->frames);
   printf("samplerate:\t%d\n",     sfinfo->samplerate);
   printf("channels:\t%d\n",       sfinfo->channels);
   printf("format number:\t%d\n",  sfinfo->format);
   printf("sections:\t%d\n",       sfinfo->sections);
   printf("seekable?:\t%d\n",      sfinfo->seekable);
   printf("\n");
}

/*
  remind the user what command line arguments we support
*/
void exitDisplayingUsage(char *command)
{
   // display usage and exit with returncode 1
   printf("usage: %s -w <wav filename> -i <img prefix> -d <img width>\n", command);
   exit(1);
}

/*
	compute amplitude envelope

  this currently does very little 
  - but the structure of sndData->ampEnv->imgData
  means we could insert non-linear processing here
  (e.g. down/upsampling, frequency equalisation, amplitude dynamics)
*/
void calcAmpEnv(
  t_config *config, 
  unsigned char *sndData, 
  int *ampEnvelope)
{
  // don't allow attenuation
  if (config->gain < 1) config->gain = 1;

  int samplesToProcess = config->sndSamplerate;
  float sndVal = 0;
	while(samplesToProcess--)
	{
    // pull the sound data out, gain and clip
    sndVal = *sndData;
    sndVal -= 128;
    sndVal *= config->gain;
    sndVal += 128;
    if (sndVal > 255) sndVal = 255;
    if (sndVal < 0) sndVal = 0;

    // store envelope
    *ampEnvelope = sndVal;

    // move forward one step in both buffers
    ampEnvelope++;
    sndData++;
  }
  if (gDebug) printf("filled envelope buffer\n");
}

/* 
  for debugging - dump amp envelope values
*/
void displayAmpEnv(t_config *config, int *ampEnvelope)
{
  int ampIdx;
  for(ampIdx = 0; ampIdx < config->imgHeight; ampIdx++)
    printf("%d\n", ampEnvelope[ampIdx]);
}

/*
  imgData represents a bitplane - dimensions 'imgWidth'x'imgHeight'  
  we draw an amplitude envelope by setting pixels to black(0) or white(255)

  envelopes are drawn in a similar way to thie ASCII diagram below
  where '#' represents a black pixel, and whitespace represents white pixel
  numbers down the left represent index values into the imgData array

  0 #####
 10 #######
 20 ##
 30 ###
 40 ####

 we avoid line-drawing and x,y co-ords for current pixel being 'drawn'
 e.g. int imgXPos = 0; 
 e.g. int imgYPos = 0; 
    
 instead, we take advantage of the fact that the data is in one big block
 this way, we calculate the number of black pixels followed by white pixels
 to draw for any given row (assuming each row is 'imgWidth' wide)

*/
void plotImageData(t_config *config, int *ampEnvelope, unsigned char *imgData)
{
  int numXDarkPixels;
  int numXLightPixels;
  int numYPixelsToProcess = config->imgHeight;
  int ampVal; // value pulled from the envelope

  // generate image data
  while(numYPixelsToProcess--)
  {
    // pull out amplitude value and
    // rescale from 0 to 255 ---> 0 to imgWidth
    ampVal = (int)((*ampEnvelope / 255.0) * config->imgWidth);

    // clip
    if (ampVal > config->imgWidth) ampVal = config->imgWidth;
    if (ampVal < 0 ) ampVal = 0;

    // calculate the number of black-followed-by white pixels
    // to fit into the image width
    numXDarkPixels = ampVal;
    numXLightPixels = config->imgWidth - numXDarkPixels;

    //printf("%04d %04d\n", numXDarkPixels, numXLightPixels);

    // draw a row of pixels
    while(numXDarkPixels--){  *imgData = 0; imgData++; }
    while(numXLightPixels--){ *imgData = 255; imgData++; }
    ampEnvelope++;
  }
  if (gDebug) printf("finished drawing image data\n");
}

// write data to disk in PGM (portable grey map) binary format
void writeImageData(t_config *config, unsigned char *imgData, char* filename)
{
  FILE *outfile = NULL;

  // open pgm file for writing
  // abort if the output file cannot be opened for writing
  outfile = fopen(filename, "wb");
  if (!outfile)
  {
    printf("==Error, cannot open output pgm for writing==\n");
    return;
  }
  if (gDebug) printf("pgm image file open for writing ok\n");

  // write header
  fprintf(outfile, "P5\n");
  fprintf(outfile, "%d %d\n", config->imgWidth, config->imgHeight);
  fprintf(outfile, "255\n");

  // write raster
  int pixelsToProcess = config->imgWidth * config->imgHeight;
  int success = fwrite(imgData, 1, pixelsToProcess, outfile);

  fclose(outfile);
  if (success) printf("wrote %s ok\n", filename);
}

